
class Constants(object):
    """ Create objects with read-only (constant) attributes. """

    def __init__(self, *args, **kwargs):
        self._d = dict(*args, **kwargs)

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __getattr__(self, key):
        return self._d[key]

    def __setattr__(self, key, value):
        if (key[0] == '_'):
            super(Constants, self).__setattr__(key, value)
        else:
            raise ValueError("setattr locked", self)

ZERO_OP = 0
MACS = Constants(
    ADD = 1,
    EXP = 16,
    LOG = 16,
    SQRT = 4,
    POW = 32,
    MUL = 1,
    DIV = 2,
    CMP = 1,
    SIN = 14,
    COS = 14,
)

METRICS = Constants(
    FLOPs = 0,
    MACs = 1,
)


import onnx

from .hooks.common.functions import (transform_to_ndarray, get_attribute, construct_volume)


def set_inputs(graph: onnx.GraphProto, dynamic_tensors: dict, tensor_map: dict) -> onnx.GraphProto:
    """ 
    Sets values for dynamic graph inputs.
    
    Args:
        graph (onnx.GraphProto): graph to set inputs for
        dynamic_tensors (dict): dictionary of dynamic tensors
        tensor_map (dict): dictionary of tensors
    Returns:
        Updated graph
    """
    for x in graph.input:
        if dynamic_tensors.keys().__contains__(x.name):
            tensor_map[x.name] = dynamic_tensors[x.name]
            dim = x.type.tensor_type.shape.dim
            for nb, dnb in zip(dim, dynamic_tensors[x.name].shape):
                nb.dim_value = dnb
    return graph


def add_outputs(graph: onnx.GraphProto, output_names: list, tensor_map: dict) -> onnx.GraphProto:
    """
    Creates new outputs for the given graph.

    Args:
        graph (onnx.GraphProto): graph to add outputs to.
        output_names (list): names of the outputs to add to the graph.
        tensor_map (dict): dictionary of tensors
    Returns:
        Updated graph
    """
    for name in output_names:
        if tensor_map is not None and name in tensor_map:
            new_output = onnx.helper.make_tensor_value_info(name, onnx.TensorProto.FLOAT, tensor_map[name].shape)
        else:
            new_output = onnx.helper.make_tensor_value_info(name, onnx.TensorProto.FLOAT, ())
        graph.output.append(new_output)
    return graph


def update_statics(graph: onnx.GraphProto, tensor_map: dict, params_map: dict) -> int:
    """
    Updates static inputs of the given graph.

    Args:
        graph (onnx.GraphProto): graph to update statics for
        tensor_map (dict): dictionary of tensors
        params_map (dict): dictionary of parameters
    Returns:
        Total amount of parameters present in the graph
    """
    total_params = 0
    for x in graph.initializer:
        ndarray = transform_to_ndarray(x)
        tensor_map[x.name] = ndarray
    for n in graph.node:
        if n.op_type == 'Constant':
            for a in n.attribute:
                if a.name == 'value':
                    tensor_map[n.output[0]] = get_attribute(a)
    for k in tensor_map:
        params_map[k] = construct_volume(tensor_map[k].shape)
        total_params += params_map[k]
    return total_params


def remove_unused_tensors(graph : onnx.GraphProto) -> None:
    """
    Removes unused tensors from the graph.

    Args:
        graph (onnx.GraphProto): graph to remove unused tensors from
    """
    consumer = {}
    producer = {initial.name: 0 for initial in graph.initializer}
    for node in graph.node:
        for input in node.input:
            consumer[input] = 0
        for output in node.output:
            producer[output] = 0
    inputs = [key for key in consumer if key not in producer]
    outputs = [key for key in producer if key not in consumer]
    valid_inputs = [input for input in graph.input if input.name in inputs]
    valid_outputs = [output for output in graph.output if output.name in outputs]
    graph.ClearField('input')
    for input in valid_inputs:
        graph.input.append(input)
    graph.ClearField('output')
    for output in valid_outputs:
        graph.output.append(output)


from .hooks.elementwise import *
from .hooks.pointwise import *
from .hooks.quantize import *
from .hooks.fusion import *


SUPPORTED_OPERATIONS = {
    'Conv' : Conv,
    'ConvTranspose' : ConvTranspose,
    'Unsqueeze' : Unsqueeze,
    'Squeeze' : Squeeze,
    'Resize' : Resize,
    'ScatterND' : ScatterND,
    'ScatterElements' : ScatterElements,
    'Pool' : Pool,
    'MaxPool' : MaxPool,
    'AveragePool' : AveragePool,
    'GlobalAveragePool' : GlobalAveragePool,
    'Gather' : Gather,
    'Constant' : Constant,
    'Concat' : Concat,
    'Reshape' : Reshape,
    'ArgMax' : ArgMax,
    'Expand' : Expand,
    'Tile' : Tile,
    'GRU' : GRU,
    'TopK' : TopK,
    'LSTM' : LSTM,
    'Compress' : Compress,
    'RoiAlign' : RoiAlign,
    'Gemm' : Gemm,
    'MatMul' : MatMul,
    'MatMulInteger' : MatMulInteger,
    'Shape' : Shape,
    'OneHot' : OneHot,
    'Pad' : Pad,
    'Split' : Split,
    'ReduceMean' : ReduceMean,
    'ReduceProd' : ReduceProd,
    'ReduceL2' : ReduceL2,
    'ReduceMax' : ReduceMax,
    'ReduceMin' : ReduceMin,
    'ReduceSum' : ReduceSum,
    'NonZero' : NonZero,
    'Transpose' : Transpose,
    'ConstantOfShape' : ConstantOfShape,
    'Flatten' : Flatten,
    'Einsum' : Einsum,
    'Less' : Less,
    'LessOrEqual' : LessOrEqual,
    'Not' : Not,
    'And' : And,
    'Where' : Where,
    'Greater' : Greater,
    'Equal' : Equal,
    'CategoryMapper' : CategoryMapper,
    'LRN' : LocalResponseNormalization,
    'Add' : Add,
    'Sub' : Sub,
    'Mul' : Mul,
    'Sum' : Sum,
    'Abs' : Abs,
    'Neg' : Neg,
    'Min' : Min,
    'Max' : Max,
    'Pow' : Pow,
    'Exp' : Exp,
    'Log' : Log,
    'Sin' : Sin,
    'Cos' : Cos,
    'Div' : Div,
    'Hardmax' : Hardmax,
    'Relu' : Relu,
    'PRelu' : PRelu,
    'Clip' : Clip,
    'Relu6' : Relu6,
    'CumSum' : CumSum,
    'Sigmoid' : Sigmoid,
    'Softmax' : Softmax,
    'Tanh' : Tanh,
    'Sqrt' : Sqrt,
    'InstanceNormalization' : InstanceNormalization,
    'Range' : Range,
    'Reciprocal' : Reciprocal,
    'HardSigmoid' : HardSigmoid,
    'LeakyRelu' : LeakyRelu,
    'Identity' : Identity,
    'Erf' : Erf,
    'Dropout' : Dropout,
    'BatchNormalization' : BatchNormalization,
    'Slice' : Slice,
    'Cast' : Cast,
    'Floor' : Floor,
    'Ceil' : Ceil,
    'DequantizeLinear' : DequantizeLinear,
    'QuantizeLinear' : QuantizeLinear,
    'QuantizeLinearMatMul' : QuantizeLinearMatMul,
    'QuantizeLinearConv' : QuantizeLinearConv,
}


import numpy as np

from .common.utilities import *
from .common.functions import *

from .elementwise import ElementwiseBase

from copy import deepcopy
from collections import defaultdict


class FusionBase(ElementwiseBase):
    """ Defines parent class for fused module hooks. """

    def __init__(self, proto):
        super().__init__(proto)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [inputs[0]]


class Identity(FusionBase):
    """ Defines identity module hook. """
    pass


class Erf(FusionBase):
    """ Defines erf module hook. """
    pass


class Dropout(FusionBase):
    """ Defines dropout module hook. """
    pass


class BatchNormalization(FusionBase):
    """ Defines batch_normalization module hook. """
    pass


class Slice(FusionBase):
    """ Defines slice module hook. """
    
    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['axes', 'end_position', 'start_position']
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        if len(inputs) == 1:
            data = inputs[0]
            start_position = self.start_position
            end_position = self.end_position
            axes = self.axes
            steps = defaultdict(lambda: 1)
        elif len(inputs) == 3:
            data, start_position, end_position = inputs[:3]
            return [data[start_position[0]:end_position[0]]]
        elif len(inputs) == 4:
            data, start_position, end_position, axes = inputs[:4]
            steps = defaultdict(lambda: 1)
        elif len(inputs) == 5:
            data, start_position, end_position, axes, steps = inputs[:5]

        index, storage = (0, deepcopy(data))
        for i in range(len(storage.shape)):
            if i in axes:
                if i == 0:
                    storage = storage[start_position[index]:end_position[index]:steps[index], ...]
                elif i == 1:
                    storage = storage[:, start_position[index]:end_position[index]:steps[index], ...]
                elif i == 2:
                    storage = storage[:, :, start_position[index]:end_position[index]:steps[index], ...]
                elif i == 3:
                    storage = storage[:, :, :, start_position[index]:end_position[index]:steps[index], ...]
                elif i == 4:
                    storage = storage[:, :, :, :, start_position[index]:end_position[index]:steps[index], ...]
                index += 1
        return [storage]


class Cast(FusionBase):
    """ Defines cast module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [inputs[0]]


class Floor(FusionBase):
    """ Defines floor module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.floor(inputs[0])]


class Ceil(FusionBase):
    """ Defines ceil module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.ceil(inputs[0])]


import numpy as np

from .hooks.common.functions import construct_volume
from .hooks.common.utilities import calculate_sparsity, zero_flag

class SparsitySearch(object):
    def __init__(self) -> None:
        pass

    def search(self, tensors: dict, threshold_size: int = 128, threshold_ratio: float = 0.40) -> dict:
        """
        Searchs for the sparsity ratio for each tensor in tensors.

        Args:
            tensors (dict): A dictionary of tensors.
            threshold_size (int): The size of the threshold. Defaults to 128.
            threshold_ratio (float): Exit criterion
        Returns:
            Information about the sparsity of each tensor.
        """

        sparse_block_map = {}

        for key, ndarray in tensors.items():
            if (construct_volume(ndarray.shape) < threshold_size):
                continue
            ratio = calculate_sparsity(ndarray)
            if ratio is None or ratio < threshold_ratio:
                continue
            block_size, block_ratio = self._search_sparse_block_size(ndarray, ratio)
            sparse_block_map[key] = {
                "block_size" : block_size,
                "block_ratio" : block_ratio,
                "ratio" : ratio,
            }
        return sparse_block_map
    
    def _search_sparse_block_size(self, ndarray: np.ndarray, ratio: float, threshold: float = 0.099) -> tuple:
        if len(ndarray.shape) == 2:
            return self._search_sparse_block_size_1d(ndarray, ratio, threshold)    
        elif len(ndarray.shape) == 4:
            return self._search_sparse_block_size_nd(ndarray, ratio, threshold)
        return ((1, 1), ratio)
    
    def _validate_block_size_1d(self, ndarray: np.ndarray, x: bool, size: int, ratio: float, 
            axis: int, threshold: float = 0.099) -> tuple:
        if x and ndarray.shape[axis] % size == 0:
            if axis == 0:
                flag = zero_flag(ndarray.reshape(-1, size, ndarray.shape[1]))
                _sum = np.sum(flag, 1)
            elif axis == 1:
                flag = zero_flag(ndarray.reshape(ndarray.shape[0], -1, size))
                _sum = np.sum(flag, -1)
            _ratio_x = (_sum == size).sum() / _sum.size
            return (True, _ratio_x) if _ratio_x > ratio - threshold else (False, ratio)
        else:
            return (False, ratio)
    
    def _validate_block_size_nd(self, ndarray: np.ndarray, x: bool, size: int, ratio: float,
            axis: int, threshold: float = 0.099):
        if x and ndarray.shape[axis] % size == 0:
            if axis == 0:
                flag = zero_flag(ndarray.reshape(-1, size, *ndarray.shape[1:]))
                _sum = np.sum(flag, 2)
            elif axis == 1:
                flag = zero_flag(ndarray.reshape(ndarray.shape[0], -1, size, *ndarray.shape[2:]))
                _sum = np.sum(flag, 1)
            _ratio_x = (_sum == size).sum() / _sum.size
            return (True, _ratio_x) if _ratio_x > ratio - threshold else (False, ratio)
        else:
            return (False, ratio)

    def _search_sparse_block_size_1d(self, ndarray: np.ndarray, ratio: float, threshold: float = 0.099) -> tuple:
        init_size, valid_size = 2, 1
        p, q = True, True
        valid_ratio = ratio
        while True:
            is_valid_q, valid_ratio = self._validate_block_size_1d(ndarray=ndarray, x=q, 
                size=init_size, axis=1, ratio=valid_ratio, threshold=threshold)

            is_valid_p, valid_ratio = self._validate_block_size_1d(ndarray=ndarray, x=p, 
                size=init_size, axis=0, ratio=valid_ratio, threshold=threshold)

            if not is_valid_q and not is_valid_p:
                break

            valid_size = init_size
            init_size *= 2
            p = is_valid_p
            q = is_valid_q

        if q and p:
            temp = ndarray.reshape(ndarray.shape[0] // valid_size, valid_size,
                ndarray.shape[1] // valid_size, valid_size)
            flag = zero_flag(temp)
            _sum = np.sum(flag, axis=(1, -1))
            _ratio_s = (_sum == (valid_size * valid_size)).sum() / _sum.size
            if _ratio_s > ratio - threshold:
                return ((valid_size, valid_size), _ratio_s)

        return ((valid_size if p else 1, valid_size if q else 1), valid_ratio)

    def _search_sparse_block_size_nd(self, ndarray: np.ndarray, ratio: float, threshold: float = 0.099) -> tuple:
        init_size, valid_size = (2, 1)
        p, q = (True, True)
        valid_ratio_p, valid_ratio_q = (ratio, ratio)
        
        while True:
            is_valid_q, valid_ratio_q = self._validate_block_size_nd(ndarray=ndarray, x=q, 
                size=init_size, axis=1, ratio=valid_ratio_q, threshold=threshold)

            is_valid_p, valid_ratio_p = self._validate_block_size_nd(ndarray=ndarray, x=p, 
                size=init_size, axis=0, ratio=valid_ratio_p, threshold=threshold)

            if not is_valid_q and not is_valid_p:
                break

            valid_size = init_size
            init_size *= 2
            p = is_valid_p
            q = is_valid_q

        if valid_size > 1 and q and p:
            temp = ndarray.reshape(ndarray.shape[0] // valid_size, valid_size, 
                ndarray.shape[1] // valid_size, valid_size, *ndarray.shape[2:])
            
            flag = zero_flag(temp)
            _sum = np.sum(flag, axis=(1, 3))
            ratios = (_sum == (valid_size * valid_size)).sum() / _sum.size
            
            if ratios > ratio - threshold:
                return ((valid_size, valid_size), ratios)
        
        if valid_ratio_p > valid_ratio_q:
            return (valid_size, 1), valid_ratio_p
        
        return ((1, valid_size), valid_ratio_q)


import numpy as np
import math

from .functions import construct_volume


def translate_axes(size, axes):
    return [axis if axis > 0 else size + axis for axis in axes]


def calculate_max_shape(shapes: list):
    max_volume = construct_volume(shapes[0])
    max_shape = tuple()
    
    for shape in shapes:
        tmp_volume = construct_volume(shape)
        if tmp_volume > max_volume:
            max_shape = shape
            max_volume = tmp_volume
    return max_shape


def calculate_max_shape_ndarray(mndarray: list[np.ndarray]):
    max_shape = mndarray[0].shape
    max_volume = construct_volume(max_shape)
    for ndarray in mndarray:
        tmp_volume = construct_volume(ndarray.shape)
        if tmp_volume > max_volume:
            max_shape = ndarray.shape
            max_volume = tmp_volume
        elif tmp_volume == max_volume:
            if len(ndarray.shape) > len(max_shape):
                max_shape = ndarray.shape
    
    return max_shape


def calculate_conv_output_shape(input, padding, kernel_size, stride, dilation):
    return int((input + padding - dilation * (kernel_size - 1) - 1) / stride + 1)


def calculate_convtranspose_output_shape(input, output_padding, padding, kernel_size, stride, dilation):
    return stride * (input - 1) + output_padding + ((kernel_size - 1) * dilation + 1) - padding


def calculate_pooling_shape(input_shape, padding, kernel_shape, dilation, stride, use_ceil = True):
    output_shape = (input_shape + padding - ((kernel_shape - 1) * dilation + 1)) / stride + 1
    return math.ceil(output_shape) if use_ceil else math.floor(output_shape)


def calculate_sparsity(ndarray):
    if len(ndarray.shape) not in [2, 4]:
        return None
    if ndarray.dtype in [np.float32, np.float64, np.int32, np.int8]:
        flag = ndarray == 0
        return flag.sum() / ndarray.size
    if ndarray.dtype == np.uint8:
        flag = ndarray == 128
        return flag.sum() / ndarray.size


def zero_flag(ndarray):
    if ndarray.dtype in [np.float32, np.float64, np.int32, np.int8]:
        return ndarray == 0
    if ndarray.dtype == np.uint8:
        return ndarray == 128



import numpy as np
import onnx

from typing import Union
from os.path import abspath, join

from .commons import (Registry, AttributeDict, SimpleTimer, profile_to_console)
from .hooks.common.functions import (construct_volume, get_tensor_shape, validate_ndarray)
from .hooks.common.constants import METRICS

from .tensors import (remove_unused_tensors, set_inputs, update_statics)
from .operations import SUPPORTED_OPERATIONS
from .sparsity import SparsitySearch

import logging
import logging.config

logging.config.fileConfig(join(abspath("."), 'logging.conf'))
logger = logging.getLogger('profiler')


class Profiler(object):
    def __init__(self) -> None:
        self.registry_instance = Registry()
        self.crawler = SparsitySearch()
        
        for op in SUPPORTED_OPERATIONS:
            self.registry_instance.register(SUPPORTED_OPERATIONS[op], op)

        self.globals = AttributeDict()
        self.globals.tensor_map = {}
        self.globals.params_map = {}
        self.globals.sparse_map = {}
        self.globals.node_map = {}
        self.globals.shared_nodes = {}

        self.measurements = AttributeDict()
        self.measurements.macs = 0.0
        self.measurements.params = 0
        self.measurements.memory = 0

    def profile(self, config: dict = None, model : Union[str, onnx.ModelProto] = None, dynamic_inputs : dict = None, 
            hidden_operations : dict = None, stdout : bool = False, verbose: bool = False) -> Union[tuple, dict]:
        if config is not None:
            return self._profile_models_from_config(hidden_operations=hidden_operations, 
                verbose=verbose, config=config, stdout=stdout)
        elif model is not None and dynamic_inputs is not None:
            return self._profile_single_model(hidden_operations=hidden_operations, 
                dynamic_inputs=dynamic_inputs, verbose=verbose, stdout=stdout, model=model)
        raise ValueError("model and config arguments cannot be None at the same time!")                               # TODO

    def _profile_models_from_config(self, config: dict, hidden_operations: dict = None, stdout: bool = False, 
            verbose: bool = False) -> dict:
        if config is None:
            raise ValueError("config cannot be None!")                                                              # TODO
        response = {}
        for e in config:
            dynamic_inputs = {x: np.zeros(shape=x["shape"], dtype=x["dtype"]) for x in config[e]["dynamic_inputs"]}
            response[e] = self._profile_single_model(
                model=config[e]["path"],
                dynamic_inputs=dynamic_inputs,
                hidden_operations=hidden_operations,
                stdout=stdout,
                verbose=verbose
            )
        return response

    def _profile_single_model(self, model : Union[str, onnx.ModelProto], dynamic_inputs : dict, 
            hidden_operations : dict = None, stdout : bool = False, verbose: bool = False):
        if isinstance(model, str):
            model = onnx.load(model)
        
        if isinstance(model, onnx.ModelProto):
            remove_unused_tensors(model.graph)
            self._profile(model.graph, dynamic_inputs, hidden_operations, verbose)
            if stdout: 
                profile_to_console(
                    messurements={
                        "macs": self.measurements.macs,
                        "memory": self.measurements.memory,
                        "params": self.measurements.params,
                    }, 
                    maps={
                        "sparse": self.globals.sparse_map,
                        "node": self.globals.node_map,
                    },
                    operation_metric=METRICS.MACs 
                )
            return (self.measurements.macs, self.measurements.params)

    def _profile_single_node(self, proto: onnx.NodeProto, inputs: list[np.ndarray], 
            outputs: list[np.ndarray]) -> tuple:
        node_class = self.registry_instance.lookup(proto.op_type)
        if node_class is not None:
            worker = node_class(proto)
            return worker.profile(inputs, outputs)
        raise NotImplementedError()                                                         # TODO

    def _retrieve_single_shape(self, proto: onnx.NodeProto, inputs: list[np.ndarray]) -> tuple:
        node_class = self.registry_instance.lookup(proto.op_type)
        if node_class is not None:
            worker = node_class(proto)
            return worker.inference_shape(inputs)
        raise NotImplementedError()                                                         # TODO

    def _profile_inputs(self, graph: onnx.GraphProto) -> None:
        for _input in graph.input:
            tensor = self.globals.tensor_map[_input.name]
            memory = construct_volume(tensor.shape) * 4
            node_data = {
                "macs" : 0,
                "params": 0,
                "memory": memory,
                "input_shape": tensor.shape,
                "output_shape": tensor.shape,
                "block_size" : (1, 1),
                "ratio" : 0, 
                "block_ratio": 0
            }
            self.globals.node_map[_input.name] = node_data
            self.measurements.memory += memory

    def _profile_sparse_node(self, node: onnx.NodeProto, node_data: dict) -> None:
        is_sparse = False
        for input in node.input:
            if input == '':
                continue
            if input in self.globals.sparse_map.keys():
                node_data |= self.globals.sparse_map[input]
                is_sparse = True
                break
        if not is_sparse:
            node_data.update(
                {
                    'block_size': (1, 1), 
                    'block_ratio': 0,
                    'ratio': 0, 
                }
            )

    def _collect_node_inputs(self, node: onnx.NodeProto, flags: dict) -> tuple:
        memory, params, inputs = (0, 0, [])
        for _input in node.input:
                if _input == '':
                    continue
                inputs.append(self.globals.tensor_map[_input])
                if _input in self.globals.params_map.keys():
                    if flags[_input] == 0:
                        params += self.globals.params_map[_input]
                        memory += self.globals.params_map[_input]
                    flags[_input] += 1
        return (memory, params, inputs)

    def _collect_node_outputs(self, node: onnx.NodeProto, input_memory: int) -> tuple:
        memory, outputs = (input_memory, [])
        for _output in node.output:
            if self.globals.tensor_map.keys().__contains__(_output):
                outputs.append(self.globals.tensor_map[_output])
                if node.op_type == 'Constant':
                    continue
                memory += construct_volume(self.globals.tensor_map[_output].shape)
        return (memory, outputs)

    def _profile_nodes(self, graph: onnx.GraphProto, sparse_model: bool, 
            hidden_operations: dict, flags: dict) -> None:
        profile_position = 0
        for _node in graph.node:
            if hidden_operations is not None and _node.op_type in hidden_operations:
                continue
            
            memory, params, inputs = self._collect_node_inputs(_node, flags)
            memory, outputs = self._collect_node_outputs(_node, memory)
            macs, _ = self._profile_single_node(_node, inputs, outputs)
            output_shape, input_shape = ((0,), (0,))
            
            if outputs:
                output_shape = outputs[0].shape
                output_shape = (0,) if len(output_shape) == 0 else output_shape
            
            if inputs:
                input_shape = inputs[0].shape
                input_shape = (0,) if len(input_shape) == 0 else input_shape
            
            if len(_node.name) == 0:
                _node.name = f'{_node.op_type}_{profile_position}'

            profile_position += 1
            memory *= 4

            node_data = {'macs': macs, 'params': params, 'memory': memory, 
                'input_shape': input_shape, 'output_shape': output_shape}

            if sparse_model:
                self._profile_sparse_node(_node, node_data)

            self.globals.node_map.update({_node.name: node_data})
            self.measurements.macs += macs
            self.measurements.params += params
            self.measurements.memory += memory

    def _profile(self, graph: onnx.GraphProto, dynamic_inputs: dict = None, 
            hidden_operations: dict = None, verbose: bool = False) -> None:
        if hidden_operations is None:
            hidden_operations = {'Identity', 'Constant'}
        internal_timer = SimpleTimer()
        self._retrieve_single_shapes(graph, dynamic_inputs)

        if verbose:
            logger.info(f"retrivial of all tensor shapes took {internal_timer.stop():3f} sec's")

        sparse_model = len(self.globals.sparse_map.keys()) > 0
        params_flags = {key: False for key in self.globals.params_map.keys()}

        self._profile_inputs(graph)
        self._profile_nodes(graph, sparse_model, hidden_operations, params_flags)

        if verbose:
            logger.info(f"profiling of all nodes took {internal_timer.stop():.3f} sec's")

        for node in graph.node:
            for _input in node.input:
                if _input == '':
                    continue
                if _input in self.globals.params_map.keys() and params_flags[_input] > 1 \
                             and construct_volume(self.globals.tensor_map[_input].shape) > 128:
                    if _input in self.globals.shared_nodes:
                        self.globals.shared_nodes[_input].append(node.name)
                    else:
                        self.globals.shared_nodes[_input] = [node.name]

        if verbose:
            count = sum(construct_volume(self.globals.tensor_map[t].shape) for t in self.globals.tensor_map) * 4
            difference = abs(self.measurements.memory - count) / count
            logger.info(f"globals -> tensor_map: {count}")
            logger.info(f"globals -> nodememory: {self.measurements.memory}")
            logger.info(f"globals -> difference: {difference:.3%}")
            assert (difference < 0.01), f"diference to large! delta: {difference}"

    def _retrieve_single_shapes(self, graph: onnx.GraphProto, dynamic_inputs: dict) -> None:
        self.globals.tensor_map = {}
        self.globals.params_map = {}

        update_statics(graph, self.globals.tensor_map, self.globals.params_map)
        self.globals.sparse_block_map = self.crawler.search(self.globals.tensor_map)

        if dynamic_inputs is not None:
            set_inputs(graph, dynamic_inputs, self.globals.tensor_map)

        for _input in graph.input:
            shape = get_tensor_shape(_input)
            for dim in shape:
                if dim < 0:
                    raise ValueError()                                                          # TODO
            if _input.name not in self.globals.tensor_map:
                self.globals.tensor_map[_input.name] = np.zeros(shape, dtype=np.float32)

            if not validate_ndarray(self.globals.tensor_map[_input.name]):
                raise ValueError()                                                              # TODO

        for node in graph.node:
            inputs = [self.globals.tensor_map[x] for x in node.input if x != '']
            outputs = [y for y in node.output if y != '']
            output_tensors = self._retrieve_single_shape(node, inputs)
            for tensor, name in zip(output_tensors, outputs):
                self.globals.tensor_map[name] = tensor

        for key in self.globals.tensor_map:
            shape = self.globals.tensor_map[key].shape
            if len(shape) == 0:
                shape = (0,)
            value_info = onnx.helper.make_tensor_value_info(key, onnx.TensorProto.FLOAT, shape)
            graph.value_info.append(value_info)

        for output in graph.output:
            dim = output.type.tensor_type.shape.dim
            for nb, dnb in zip(dim, self.globals.tensor_map[output.name].shape):
                nb.dim_value = dnb


from time import time
from tabulate import tabulate

from .hooks.common.constants import METRICS


class Registry(object): 
    def __init__(self):
        self.__registry = {}
    
    def __setitem__(self, key, obj):
        if key in self.__registry:
            raise KeyError(f"'{key}' was already registered!")
        
        self.__registry[key] = obj
    
    def __getitem__(self, key):
        if self.__registry.__contains__(key) is False:
            raise KeyError(f"entry '{key}' was not found!")
        
        return self.__registry[key]
    
    def __contains__(self, key):
        return key in self.__registry
    
    def __iter__(self):
        return iter(self.__registry.items())
    
    def register(self, obj = None, name = None):
        """ Registers a object under a given name.
        Can be used as decorator or method.
    
        Args:
            obj: The object to add to the registry.
            name: The name of the object.
        Raises:
            KeyError: If same name is used twice.
        """
        if obj is None:
            def dec(x):
                _name = x.__name__ if name is None else name
                self.__setitem__(_name, x)
                return x
            return dec

        _name = obj.__name__ if name is None else name
        self.__setitem__(_name, obj)
    
    def lookup(self, key):
        """ Looks up "key".
    
        Args:
            key: a string specifying the registry key for the object.
        Returns:
            Registered object if found
        Raises:
            KeyError: if "name" has not been registered.
        """
        if key in self.__registry:
            return self.__registry[key]
        else:
           raise KeyError(f"entry '{key}' was not found!")
    
    def list(self):
        """ Lists registered objects.
        
        Returns:
            A list of names of registered objects.
        """
        return self.__registry.keys()


class AttributeDict(dict):
    IMMUTABLE = "__immutable__"

    def __init__(self, *args, **kwargs):
        super(AttributeDict, self).__init__(*args, **kwargs)
        self.__dict__[AttributeDict.IMMUTABLE] = False

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        elif name in self:
            return self[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if self.__dict__[AttributeDict.IMMUTABLE]:
            raise AttributeError(f"Attempted to set '{name}' to '{value}', but AttrDict is immutable")
        if name in self.__dict__:
            self.__dict__[name] = value
        else:
            self[name] = value

    def immutable(self, is_immutable):
        """
        Set immutability to is_immutable and recursively apply the setting
        to all nested AttributeDicts.

        Args:
            is_immutable: controls IMMUTABLE parameter.
        """
        self.__dict__[AttributeDict.IMMUTABLE] = is_immutable
        # Recursively set immutable state
        for v in self.__dict__.values():
            if isinstance(v, AttributeDict):
                v.immutable(is_immutable)
        for v in self.values():
            if isinstance(v, AttributeDict):
                v.immutable(is_immutable)

    def is_immutable(self):
        return self.__dict__[AttributeDict.IMMUTABLE]


class SimpleTimer():
    def __init__(self):
        self.start_time = time()

    def start(self):
        """ Starts timer. """
        self.start_time = time()

    def stop(self):
        """ Stops timer. """
        return time() - self.start_time


def _tuple_to_string(x: tuple, split_character: str = ",") -> str:
    response_string = ""
    for i, value in enumerate(x):
        response_string += str(value)
        if i != len(x) - 1:
            response_string += split_character
    return response_string


def _number_to_string(x: int, use_csv_format: bool = True):
    return f"{x:,}" if use_csv_format else f"{x}"


def profile_to_console(messurements: dict, maps: dict, operation_metric: int = METRICS.FLOPs) -> None:
    assert (operation_metric in [METRICS.FLOPs, METRICS.MACs]), f"unknown metric: {operation_metric}"

    has_sparse_contents = len(maps["sparse"].keys()) != 0
    factor = 2 if operation_metric == METRICS.FLOPs else 1

    table_header = ["Name"]
    if has_sparse_contents:
        table_header.extend(("Sparse Pattern", "Sparse Block Ratio", "Sparse Ratio"))
    if operation_metric == METRICS.MACs:
        table_header.extend(("MACs", "MACs %"))
    if operation_metric == METRICS.FLOPs:
        table_header.extend(("FLOPs", "FLOPs %"))
    table_header.extend(("Memory", "Memory %", "Parameters", "Parameters %", "Input Shape", "Output Shape"))

    table_body = []
    for key, item in maps["node"].items():
        row = [key]
        if has_sparse_contents:
            row.extend((_tuple_to_string(item["block_size"], split_character="x"), 
                "{:.2%}".format(item["block_ratio"]), "{:.2%}".format(item["ratio"])))

        row.extend((_number_to_string(int(item["macs"]) * factor), "{:.2%}".format(item["macs"] / messurements["macs"])))
        row.extend((_number_to_string(int(item["memory"])), "{:.2%}".format(item["memory"] / messurements["memory"])))
        row.extend((_number_to_string(int(item["params"])), "{:.2%}".format(item["params"] / messurements["params"])))
        row.append(_tuple_to_string(item["input_shape"], split_character="x"))
        row.append(_tuple_to_string(item["output_shape"], split_character="x"))
        table_body.append(row)

    table_footer = ["TOTAL"]
    if has_sparse_contents:
        table_footer.extend(("_", "_", "_"))
    table_footer.extend((
        _number_to_string(int(messurements["macs"] * factor)), "100%", 
        _number_to_string(int(messurements["memory"])), "100%", 
        _number_to_string(int(messurements["params"])), "100%", 
        "_", "_"))

    table_body.append(table_footer)
    print(tabulate(table_body, headers=table_header, tablefmt="psql"))


import numpy as np
import argparse

from json import dumps
from typing import Union, List, Any
from .profiler import Profiler
from os.path import splitext, join, abspath


import logging
import logging.config

logging.config.fileConfig(join(abspath("."), 'logging.conf'))
logger = logging.getLogger("root")


class HelpFormatter(argparse.HelpFormatter):
    def __init__(self, prog: Any) -> None:
        super().__init__(prog, max_help_position=40, width=80)
        self._action_counter = 0

    def _format_action_invocation(self, action : argparse.Action) -> str:
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)
        
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)
        self._action_counter += 1
        
        return ", ".join(action.option_strings) + " <" + args_string.lower().split("_")[1] + ">"
 

def _retrieve_parser() -> argparse.Namespace:
    formatter = lambda prog: HelpFormatter(prog)
    parser = argparse.ArgumentParser(description="A general and accurate MACs / FLOPs profiler for ONNX models.",
        prog="onnxprofile",
        usage="%(prog)s [-h] [-v,--version] -i <model> [-d <dynamic>] [-c] [-v]",
        formatter_class=formatter,
        )
    parser.add_argument("-i", "--input", dest="_model_input_location", required=True, 
        help="specify the path to the ONNX model")
    parser.add_argument("-d", "--dynamic_inputs", nargs='+', dest="_tensor_inputs", 
        help="specify the shape of the io tensors like as follows: \"input:>f4:1x3xHxW\"")
    parser.add_argument("-c", "--console_output", action="store_true", default=False, 
        dest="_console_output", help="displays a detailed report to the console")
    parser.add_argument("-v", "--verbose", action="store_true", default=False, 
        dest="_verbose", help="displays some additional debug information")
    
    return parser.parse_args()


def _string_to_dtype(encoded_type: str) -> Union[np.float32, np.int32, np.int64]:
    return np.dtype(encoded_type)


def _string_to_list(encoded_list : str, dtype: np.dtype) -> list:
    return [int(value) for value in encoded_list.split("x")]


def _string_to_ndarray(arguments: List[str]) -> np.ndarray:
    response = {}
    for argument in arguments:
        contents = argument.split(":")
        dtype = _string_to_dtype(contents[1])
        shape = _string_to_list(contents[2], dtype=np.int32)

        if len(contents) > 3:
            ndarray = np.array(_string_to_list(contents[3], 
                dtype=dtype), dtype=dtype).reshape(shape)
        else:
            ndarray = np.zeros(shape, dtype=dtype)
        response[contents[0]] = ndarray
    return response


def main():
    arguments = _retrieve_parser()
    profiler_instance = Profiler()
    
    output_file_path = f"{splitext(arguments._model_input_location)[0]}.json"
    model_name = splitext(arguments._model_input_location)[0].split("\\")[-1]

    if arguments._tensor_inputs is not None:
        dynamic_inputs = _string_to_ndarray(arguments._tensor_inputs)
    else:
        dynamic_inputs = None
    
    if arguments._verbose: 
        logger.info(f"input -> \"{model_name}\"")
    
    macs, params = profiler_instance.profile(
        model=arguments._model_input_location,
        dynamic_inputs=dynamic_inputs, 
        stdout=arguments._console_output, 
        verbose=arguments._verbose
    )
    with open(output_file_path, "w") as stream:
        stream.write(dumps(obj={"model" : model_name, "result": {"macs": float(macs), "params": int(params)}}, indent=4))
    
    if arguments._verbose: 
        logger.info(f"output -> \"{output_file_path}\"")


if __name__ == "__main__":
    main()


import numpy as np
import math

from .common.utilities import *
from .common.functions import *
from .common.constants import MACS, ZERO_OP


class ElementwiseBase(object):
    """ Defines parent class for elementwise module hooks. """

    def __init__(self, proto):
        self.dimensions = {
            "inputs": len(proto.input),
            "outputs": len(proto.output),
        }
        self.module_name = proto.name

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        return (ZERO_OP, ZERO_OP)
    
    def inference_shape(self, inputs: list[np.ndarray]):
        return [(ZERO_OP), ]


class Conv(ElementwiseBase):
    """ Defines conv module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['pads', 'strides', 'dilations', 'auto_pad']
        self.auto_pad = None
        self.pads = (0, 0, 0, 0)
        self.strides = (1, 1)
        self.dilations = (1, 1)
        set_attributes(proto.attribute, attribute_names, self)
    
    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        if self.dimensions["outputs"] == 1 and self.dimensions["inputs"] in [2, 3]:
            kernel_shape = inputs[1].shape
            params = construct_volume(kernel_shape)
            if self.dimensions["inputs"] == 3:
                params += kernel_shape[0]

            output_volume = construct_volume(outputs[0].shape)
            if len(kernel_shape) > 3:
                macs = output_volume * kernel_shape[1] * kernel_shape[2] * kernel_shape[3]
                macs += output_volume * MACS.ADD
            elif len(kernel_shape) == 3:
                macs = output_volume * kernel_shape[1] * kernel_shape[2]
                macs += output_volume * MACS.ADD
        return (macs, params)

    def inference_shape(self, inputs: list[np.ndarray]):
        shape = []
        x_tensor, w_tensor = inputs[:2]
        x_shape = x_tensor.shape
        w_shape = w_tensor.shape
        if self.auto_pad is None or self.auto_pad == b'NOTSET':
            if len(x_shape) == 4:
                oh = calculate_conv_output_shape(
                    x_shape[2], 
                    self.pads[0] + self.pads[2], 
                    w_shape[2], 
                    self.strides[0], 
                    self.dilations[0]
                )
                ow = calculate_conv_output_shape(
                    x_shape[3], 
                    self.pads[1] + self.pads[3], 
                    w_shape[3], 
                    self.strides[1], 
                    self.dilations[1]
                )
                shape = (x_shape[0], w_shape[0], oh, ow)
            elif len(x_shape) == 3:
                oh = calculate_conv_output_shape(
                    x_shape[2], 
                    self.pads[0] + self.pads[1], 
                    w_shape[2], 
                    self.strides[0],
                    self.dilations[0]
                )
                shape = (x_shape[0], w_shape[0], oh)
        elif self.auto_pad in [b'SAME_LOWER', b'SAME_UPPER']:
            shape = (x_shape[0], w_shape[0], math.ceil(x_shape[2] / self.strides[0]))
            if len(x_shape) == 4:
                shape += (math.ceil(x_shape[3] / self.strides[1]), )
        return [construct_ndarray(shape, dtype=np.float32)]


class ConvTranspose(ElementwiseBase):
    """ Defines conv_transpose module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['pads', 'strides', 'dilations', 'output_padding', 'output_shape', 'group']
        self.pads = (0, 0, 0, 0)
        self.strides = (1, 1)
        self.dilations = (1, 1)
        self.output_padding = (0, 0, 0, 0)
        self.output_shape = (0, 0)
        self.group = 1
        set_attributes(proto.attribute, attribute_names, self)

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        if self.dimensions["outputs"] == 1 and self.dimensions["inputs"] in [3, 2]:
            kernel_shape = inputs[1].shape
            params = output_volume(kernel_shape)
            if self.dimensions["inputs"] == 3:
                params += kernel_shape[0]

            output_volume = construct_volume(inputs[0].shape)
            if len(kernel_shape) > 3:
                macs = output_volume * kernel_shape[1] * kernel_shape[2] * kernel_shape[3]
                macs += output_volume * MACS.ADD
            elif len(kernel_shape) == 3:
                macs = output_volume * kernel_shape[1] * kernel_shape[2]
                macs += output_volume * MACS.ADD
        return (macs, params)

    def inference_shape(self, inputs: list[np.ndarray]):
        x_tensor, w_tensor = inputs[:2]
        x_shape = x_tensor.shape
        w_shape = w_tensor.shape
        shape = []
        channels = self.group * w_shape[1]
        if len(x_shape) == 4:
            output_w = calculate_convtranspose_output_shape(
                x_shape[2], 
                self.output_padding[0], 
                self.pads[0] + self.pads[2], 
                w_shape[2], 
                self.strides[0], 
                self.dilations[0]
            )
            output_h = calculate_convtranspose_output_shape(
                x_shape[3], 
                self.output_padding[1],
                self.pads[1] + self.pads[3], 
                w_shape[3], 
                self.strides[1],
                self.dilations[1]
            )
            shape = [x_shape[0], channels, output_w, output_h]
            if construct_volume(self.output_shape) != 0:
                shape[2:] = self.output_shape
        elif len(x_shape) == 3:
            output_w = calculate_convtranspose_output_shape(
                x_shape[2], 
                self.output_padding[0], 
                self.pads[0] + self.pads[1],
                w_shape[2], 
                self.strides[0], 
                self.dilations[0]
            )
            shape = [x_shape[0], channels, output_w]
            if construct_volume(self.output_shape) != 0:
                shape[2] = self.output_shape[0]
        return [construct_ndarray(shape, dtype=np.float32)]


class Unsqueeze(ElementwiseBase):
    """ Defines unsqueeze module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.axes = None
        set_attributes(proto.attribute, ['axes'], self)

    def inference_shape(self, inputs: list[np.ndarray]):
        outputs = inputs[0]
        axes = inputs[1] if self.axes is None else self.axes
        for axis in axes:
            outputs = np.expand_dims(outputs, axis=axis)
        return [outputs]


class Squeeze(ElementwiseBase):
    """ Defines squeeze module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.axes = [0]
        for a in proto.attribute:
            if a.name == 'axes':
                self.axes = get_attribute(a)

    def inference_shape(self, inputs: list[np.ndarray]):
        outputs = inputs[0]
        if self.dimensions["inputs"] == 2:
            self.axes = inputs[1]
        for idx, axis in enumerate(self.axes):
            outputs = np.squeeze(outputs, axis=axis - idx)
        return [outputs]


class Resize(ElementwiseBase):
    """ Defines resize module hook. """

    def __init__(self, proto):
        attribute_names = ['mode']
        set_attributes(proto.attribute, attribute_names, self)

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        output_volume = construct_volume(outputs[0].shape)
        if self.mode == b'nearest':
            output_volume *= 0
        elif self.mode == b'linear':
            output_volume *= 4
        elif self.mode == b'cubic':
            output_volume *= 8
        return (output_volume, ZERO_OP)

    def inference_shape(self, inputs: list[np.ndarray]):
        x = inputs[0]
        sizes, output_shape = ([], [])

        if len(inputs) == 2:
            scales = inputs[1]
        elif len(inputs) >= 3:
            scales = inputs[2]
            if len(inputs) >= 4:
                sizes = inputs[3]

        if validate_ndarray(sizes):
            if len(sizes) == 4:
                output_shape = sizes
            if len(sizes) == 2:
                output_shape = x.shape[:2] + sizes
        elif validate_ndarray(scales):
            for src, scale in zip(x.shape, scales):
                output_shape.append(math.floor(src * scale))

        if validate_ndarray(output_shape) and output_shape.dtype != np.int64:
            output_shape = output_shape.astype(dtype=np.int64)

        return [construct_ndarray(output_shape, dtype=np.flaot32)]


class ScatterND(ElementwiseBase):
    """ Defines scatter_nd module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        data, indices, updates = inputs[:3]
        assert indices.shape[-1] <= len(data.shape)
        assert updates.shape == indices.shape[:-1] + data.shape[indices.shape[-1]:]

        outputs = np.copy(data)
        for i in np.ndindex(indices.shape[:-1]):
            outputs[indices[i]] = updates[i]
        return [outputs]


class Pool(ElementwiseBase):
    """ Defines pool module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['kernel_shape', 'pads', 'strides', 'ceil_mode']
        self.kernel_shape = (3, 3)
        self.ceil_mode = 0
        self.pads = (0, 0, 0, 0)
        self.strides = (1, 1)
        self.dilations = (1, 1)
        set_attributes(proto.attribute, attribute_names, self)
        self.operation_mac_count = MACS.CMP

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        output_volume = construct_volume(outputs[0].shape)
        macs = output_volume * MACS.CMP * self.kernel_shape[0]
        if len(self.kernel_shape) == 2:
            macs *= self.kernel_shape[1]
        return (macs, 0)

    def inference_shape(self, inputs: list[np.ndarray]):
        if len(self.kernel_shape) == 1:
            input_shape = inputs[0].shape
            output_shape = input_shape[:2] + (
                calculate_pooling_shape(
                    input_shape[2], 
                    self.pads[0] + self.pads[1], 
                    self.kernel_shape[0], 
                    self.dilations[0], 
                    self.strides[0], 
                    self.ceil_mode
                ),
            )
            return [construct_ndarray(output_shape, dtype=np.float32)]
        if len(self.kernel_shape) == 2:
            input_shape = inputs[0].shape
            output_shape = input_shape[:2] + (
                calculate_pooling_shape(
                    input_shape[2], 
                    self.pads[0] + self.pads[2], 
                    self.kernel_shape[0], 
                    self.dilations[0], 
                    self.strides[0], 
                    self.ceil_mode
                ),
                calculate_pooling_shape(
                    input_shape[3],
                    self.pads[1] + self.pads[3], 
                    self.kernel_shape[1], 
                    self.dilations[1], 
                    self.strides[1], 
                    self.ceil_mode
                ),
            )
            return [construct_ndarray(output_shape, dtype=np.float32)]


class MaxPool(Pool):
    """ Defines max_pool module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.CMP


class AveragePool(Pool):
    """ Defines average_pool module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.ADD


class GlobalAveragePool(ElementwiseBase):
    """ Defines global_average_pool module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        input_shape = inputs[0].shape
        shape = input_shape[:2] + tuple([1] * (len(input_shape) - 2))
        return [construct_ndarray(shape, dtype=np.float32)]

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        return (construct_volume(inputs[0].shape) * MACS.ADD + construct_volume(outputs[0].shape) * MACS.DIV, ZERO_OP)


class Gather(ElementwiseBase):
    """ Defines gather module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.axis = 0
        for a in proto.attribute:
            if a.name == 'axis':
                self.axis = a.i

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.take(inputs[0], inputs[1].astype(dtype=np.int32), axis=self.axis)]


class Constant(ElementwiseBase):
    """ Defines constant module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.value = 0
        for a in proto.attribute:
            if a.name == 'value':
                self.value = get_attribute(a)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [self.value]


class Concat(ElementwiseBase):
    """ Defines concat module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.axis = 0
        for a in proto.attribute:
            if a.name == 'axis':
                self.axis = get_attribute(a)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.concatenate(inputs, self.axis)]


class Reshape(ElementwiseBase):
    """ Defines reshape module hook. """

    def __init__(self, proto):
        super().__init__(proto)

    def inference_shape(self, inputs: list[np.ndarray]):
        input_shape = inputs[0].shape
        shape = inputs[1]
        output_shape = []
        for i in range(len(shape)):
            if shape[i] == 0:
                output_shape.append(int(input_shape[i]))
            else:
                output_shape.append(int(shape[i]))
        try:
            outputs = (inputs[0].reshape(output_shape))
        except Exception:
            outputs = np.zeros(shape.astype(np.int64), inputs[0].dtype)
        return [outputs]


class ArgMax(ElementwiseBase):
    """ Defines argmax module hook. """

    def __init__(self, n):
        super().__init__(n)
        attribute_names = ['axis', 'keepdims']
        self.keepdims = 1
        set_attributes(n.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [construct_ndarray(inputs[0], self.axis, self.keepdims)]


class Expand(ElementwiseBase):
    """ Defines argmax module hook. """

    def __init__(self, proto):
        super().__init__(proto)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [inputs[0] * np.ones(inputs[1].astype(np.int64), dtype=np.float32)]


class Tile(ElementwiseBase):
    """ Defines argmax module hook. """

    def __init__(self, proto):
        super().__init__(proto)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.tile(inputs[0], inputs[1])]


class GRU(ElementwiseBase):
    """ Defines argmax module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['hidden_size']
        set_attributes(proto.attribute, attribute_names, self)

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        w, r, b = inputs[1:4]
        params = construct_volume(w.shape) + construct_volume(r.shape) + construct_volume(b.shape)
        batch = inputs[0].shape[1]
        macs = construct_volume(w.shape) + construct_volume(r.shape) + construct_volume(b.shape) * MACS.ADD
        return (macs * batch, params)

    def inference_shape(self, inputs: list[np.ndarray]):
        x, w = inputs[:2]
        seq_len, batch = x.shape[:2]
        num_dir = w.shape[0]
        h_len = w.shape[1] // 3
        return [construct_ndarray((seq_len, num_dir, batch, h_len), dtype=np.float32), 
                construct_ndarray((num_dir, batch, h_len), dtype=np.float32)]


class TopK(ElementwiseBase):
    """ Defines argmax module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['axis']
        self.axis = None
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        x = inputs[0]
        k = inputs[1][0]
        newshape = []
        for i in range(len(x.shape)):
            if i == self.axis:
                newshape.append(k)
            else:
                newshape.append(x.shape[i])
        return [construct_ndarray(newshape, dtype=np.float32), construct_ndarray(newshape, dtype=np.int64)]


class LSTM(ElementwiseBase):
    """ Defines lstm module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['direction', 'hidden_size']
        self.direction = None
        self.hidden_size = None
        set_attributes(proto.attribute, attribute_names, self)

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        w, r, b = inputs[1:4]
        params = construct_volume(w.shape) + construct_volume(r.shape) + construct_volume(b.shape)
        batch = inputs[0].shape[1]
        macs = construct_volume(w.shape) + construct_volume(r.shape) + construct_volume(b.shape) * MACS.ADD
        return (macs * batch, params)

    def inference_shape(self, inputs: list[np.ndarray]):
        x, w = inputs[:2]
        seq_len, batch = x.shape[:2]
        num_dir = w.shape[0]
        h_len = w.shape[1] // 4
        return [construct_ndarray((seq_len, num_dir, batch, h_len), dtype=np.float32),
                construct_ndarray((num_dir, batch, h_len), dtype=np.float32)]


class Compress(ElementwiseBase):
    """ Defines compress module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['axis']
        self.axis = None
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.compress(inputs[1], inputs[0], self.axis)]


class RoiAlign(ElementwiseBase):
    """ Defines roi_align module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['mode', 'output_height', 'output_width', 'sampling_ratio', 'spatial_scale']
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        if len(inputs[0].shape) == 4 and self.output_height is not None and self.output_width is not None:
            newshape = inputs[0].shape[:2] + (self.output_height, self.output_width)
        else:
            raise NotImplementedError()
        return [construct_ndarray(newshape, dtype=np.float32)]


class ScatterElements(ElementwiseBase):
    """ Defines scatter_elements module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['axis']
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        data, indices, updates = inputs[:3]
        assert indices.shape[-1] <= len(data.shape)
        assert updates.shape == indices.shape[:-1] + data.shape[indices.shape[-1]:]

        outputs = np.copy(data)
        for i in np.ndindex(indices.shape[:-1]):
            if self.axis == 'add':
                outputs[indices[i]] += updates[i]
            elif self.axis == 'mul':
                outputs[indices[i]] *= updates[i]
            else:
                outputs[indices[i]] = updates[i]
        return [outputs]


class Gemm(ElementwiseBase):
    """ Defines gemm module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['transA', 'transB']
        self.transA = None
        self.transB = None
        set_attributes(proto.attribute, attribute_names, self)

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        x_shape = inputs[0].shape
        if self.dimensions["inputs"] < 2:
            raise NotImplementedError()
        weight_shape = inputs[1].shape
        params = construct_volume(weight_shape)
        if self.dimensions["inputs"] == 3:
            params += construct_volume(inputs[2].shape)

        macs = construct_volume(x_shape)
        if self.__class__ == Gemm:
            macs *= weight_shape[0]
        else:
            macs *= weight_shape[-1]

        if self.dimensions["inputs"] == 3:
            macs += construct_volume(outputs[0].shape) * MACS.ADD
        return (macs, params)

    def inference_shape(self, inputs: list[np.ndarray]):
        x, w = inputs[:2]

        if self.__class__ == Gemm:
            if self.transA is not None and self.transA > 0:
                x_shape = x.shape[::-1]
            else:
                x_shape = x.shape
            if self.transB is not None and self.transB > 0:
                y_shape = x_shape[:-1] + (w.shape[-2],)
            else:
                y_shape = x_shape[:-1] + (w.shape[-1],)
        else:
            y_shape = x.shape[:-1] + (w.shape[-1],)

        return [construct_ndarray(y_shape, dtype=np.float32)]


class MatMul(Gemm):
    """ Defines mat_mul module hook. """
    pass


class MatMulInteger(Gemm):
    """ Defines mat_mul_integer module hook. """
    pass


class Shape(ElementwiseBase):
    """ Defines shape module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.array(inputs[0].shape, np.int32)]


class OneHot(ElementwiseBase):
    """ Defines one_hot module hook. """    

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['axis']
        self.axis = -1
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        indices, depth = inputs[:2]
        values = np.asarray(indices)
        rank = len(values.shape)
        depth_range = np.arange(depth)
        if axis < 0:
            axis += (rank + 1)
        ls = values.shape[:self.axis]
        rs = values.shape[self.axis:rank]
        targets = np.reshape(depth_range, (1,) * len(ls) + depth_range.shape + (1,) * len(rs))
        values = np.reshape(np.mod(values, depth), ls + (1,) + rs)
        return [np.asarray(targets == values, dtype=np.float32)]


class Pad(ElementwiseBase):
    """ Defines pad module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['pads', 'value']
        self.pads = None
        self.value = 0
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        data = inputs[0]
        newshape = []
        if self.pads is None:
            if self.dimensions["inputs"] > 1:
                pads = inputs[1]
                newshape.extend(v + pads[i] + pads[i + len(data.shape)] for i, v in enumerate(data.shape))
        else:
            newshape.extend(v + self.pads[i] + self.pads[i + len(data.shape)] for i, v in enumerate(data.shape))
        return [construct_ndarray([int(val) for val in newshape], dtype=np.float32)]


class Split(ElementwiseBase):
    """ Defines split module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['axis', 'split']
        self.axis = None
        self.split = None
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        split = []
        end = 0
        if self.split is None:
            if len(inputs) == 2:
                self.split = inputs[1]
            else:
                self.split = [inputs[0].shape[self.axis] // 2]

        self.axis = translate_axes(len(inputs[0].shape), [self.axis])[0]
        for v in self.split:
            if end + v >= inputs[0].shape[self.axis]:
                break
            split.append(end + v)
            end += v
        return np.split(inputs[0], split, self.axis)


class ReduceMean(ElementwiseBase):
    """ Defines reduce_mean module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['axes', 'keepdims']
        self.axes = None
        self.keepdims = 1
        set_attributes(proto.attribute, attribute_names, self)
        self.axes = tuple(self.axes) if self.axes is not None else None

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        return (construct_volume(inputs[0].shape) * MACS.ADD, ZERO_OP)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.mean(inputs[0], axis=self.axes, keepdims=self.keepdims == 1)]


class ReduceProd(ReduceMean):
    """ Defines reduce_prod module hook. """

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        return (construct_volume(inputs[0].shape) * MACS.MUL, ZERO_OP)
    
    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.prod(inputs[0], axis=self.axes, keepdims=self.keepdims == 1)]


class ReduceL2(ElementwiseBase):
    """ Defines reduce_l2 module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['axes', 'keepdims']
        self.axes = None
        self.keepdims = 1
        set_attributes(proto.attribute, attribute_names, self)
        self.axes = tuple(self.axes) if self.axes is not None else None

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        return (construct_volume(inputs[0].shape) * (MACS.ADD + MACS.SQRT), ZERO_OP)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.sqrt(np.sum(inputs[0], axis=self.axes, keepdims=self.keepdims == 1))]


class ReduceSum(ReduceMean):
    """ Defines reduce_sum module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.sum(inputs[0], axis=self.axes, keepdims=self.keepdims == 1)]


class ReduceMin(ReduceMean):
    """ Defines reduce_min module hook. """

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        return (construct_volume(inputs[0].shape) * MACS.CMP, ZERO_OP)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.minimum.reduce(inputs[0], axis=self.axes, keepdims=self.keepdims == 1)]


class ReduceMax(ReduceMin):
    """ Defines reduce_max module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.maximum.reduce(inputs[0], axis=self.axes, keepdims=self.keepdims == 1)]


class NonZero(ElementwiseBase):
    """ Defines non_zero module hook. """

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        return (construct_volume(outputs[0].shape) * MACS.CMP, ZERO_OP)

    def inference_shape(self, inputs: list[np.ndarray]):
        result = np.array(np.nonzero(inputs[0]), dtype=np.int64)
        if construct_volume(result.shape) == 0:
            result = np.array(np.nonzero(np.ones_like(inputs[0])), dtype=np.int64)
        return [result]


class Transpose(ElementwiseBase):
    """ Defines transpose module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['perm']
        self.perm = None
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.transpose(inputs[0], self.perm)]


class ConstantOfShape(ElementwiseBase):
    """ Defines constants_of_shape module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['value']
        self.value = None
        set_attributes(proto.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        arr = np.zeros(inputs[0].astype(np.int64), dtype=np.float32)
        if self.value is not None and len(self.value) == 1:
            arr.fill(self.value[0])
        return [arr]


class Flatten(ElementwiseBase):
    """ Defines flatten module hook. """

    def __init__(self, node):
        super().__init__(node)
        attribute_names = ['axis']
        self.axis = None
        set_attributes(node.attribute, attribute_names, self)

    def inference_shape(self, inputs: list[np.ndarray]):
        if self.axis is None:
            return [inputs[0].reshape((inputs[0].shape[0], -1))]
        tmp_volume = 1
        for i in range(self.axis):
            tmp_volume *= inputs[0].shape[i]
        return [inputs[0].reshape((tmp_volume, -1))]


class Einsum(ElementwiseBase):
    """ Defines einsum module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        set_attributes(proto.attribute, ['equation'], self)
        strings = self.equation.split(b',')
        self.a_shape = strings[0].replace(b' ', b'')
        strings = strings[1].split(b'->')
        self.b_shape = strings[0].replace(b' ', b'')
        self.c_shape = strings[1].replace(b' ', b'')

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        macs = 1
        maps = {self.ashape[i]: v for i, v in enumerate(inputs[0].shape)}
        for i, v in enumerate(inputs[1].shape):
            maps[self.bshape[i]] = v
        for value in maps.values():
            macs *= value
        return (macs, ZERO_OP)

    def inference_shape(self, inputs: list[np.ndarray]):
        maps = {self.ashape[i]: v for i, v in enumerate(inputs[0].shape)}
        for i, v in enumerate(inputs[1].shape):
            maps[self.bshape[i]] = v
        shape = [maps[k] for k in self.cshape]
        return [construct_ndarray(shape, dtype=np.float32)]


class Less(ElementwiseBase):
    """ Defines less module hook. """

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        return (construct_volume(outputs[0].shape) * MACS.CMP, ZERO_OP)

    def inference_shape(self, inputs: list[np.ndarray]):
        result = np.less(inputs[0], inputs[1])
        return [result]


class LessOrEqual(Less):
    """ Defines less_or_equal module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.less_equal(inputs[0], inputs[1])]


class Not(ElementwiseBase):
    """ Defines not module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        result = np.logical_not(inputs[0].astype(np.bool))
        return [result]


class And(ElementwiseBase):
    """ Defines and module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.logical_and(inputs[0].astype(np.bool), inputs[1].astype(np.bool))]


class Where(ElementwiseBase):
    """ Defines where module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.where(inputs[0], inputs[1], inputs[2])]


class Equal(ElementwiseBase):
    """ Defines equal module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.equal(inputs[0], inputs[1])]


class Greater(ElementwiseBase):
    """ Defines greater module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.greater(inputs[0], inputs[1])]


import numpy as np

from .common.utilities import *
from .common.functions import *
from .common.constants import MACS, ZERO_OP

from .elementwise import ElementwiseBase


class PointwiseBase(ElementwiseBase):
    """ Defines parent class for pointwise module hooks. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = ZERO_OP
        self.ratio = max(1, self.dimensions["inputs"] - 1)

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        return (construct_volume(outputs[0].shape) * self.ratio * self.operation_mac_count, ZERO_OP)

    def inference_shape(self, inputs: list[np.ndarray]):
        return [construct_ndarray(calculate_max_shape_ndarray(inputs), dtype=np.float32)]


class CategoryMapper(PointwiseBase):
    """ Defines category_mapper module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = ZERO_OP


class LocalResponseNormalization(PointwiseBase):
    """ Defines local_response_normalization module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        for a in proto.attribute:
            if a.name == 'size':
                self.size = a.i

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        output_volume = construct_volume(outputs[0].shape)
        output_volume *= (MACS.DIV + MACS.EXP + MACS.ADD + self.size * MACS.MUL)
        return (output_volume, 0)


class Add(PointwiseBase):
    """ Defines add module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.ADD

    def inference_shape(self, inputs: list[np.ndarray]):
        return [inputs[0] + inputs[1]]


class Sum(PointwiseBase):
    """ Defines sum module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.ADD

    def inference_shape(self, inputs: list[np.ndarray]):
        y = inputs[0]
        for i in range(1, len(inputs)):
            y += inputs[i]
        return [y]


class Abs(PointwiseBase):
    """ Defines abs module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.ADD

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.abs(inputs[0])]


class Neg(PointwiseBase):
    """ Defines neg module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.ADD

    def inference_shape(self, inputs: list[np.ndarray]):
        return [-inputs[0]]


class Sub(PointwiseBase):
    """ Defines sub module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.ADD

    def inference_shape(self, inputs: list[np.ndarray]):
        return [inputs[0] - inputs[1]]


class Min(PointwiseBase):
    """ Defines min module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.CMP
        self.ratio = self.dimensions["inputs"] - 1

    def inference_shape(self, inputs: list[np.ndarray]):
        result = inputs[0]
        for i in range(1, self.dimensions["inputs"]):
            result = np.minimum(result, inputs[i])
        return [result]


class Max(Min):
    """ Defines max module hook. """

    def inference_shape(self, inputs: list[np.ndarray]):
        result = inputs[0]
        for i in range(1, self.dimensions["inputs"]):
            result = np.maximum(result, inputs[i])
        return [result]


class Hardmax(PointwiseBase):
    """ Defines hardmax module hook. """
    pass


class Relu(PointwiseBase):
    """ Defines relu module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.CMP
        self.ratio = 1


class PRelu(PointwiseBase):
    """ Defines prelu module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.CMP + MACS.MUL
        self.ratio = 1


class Clip(PointwiseBase):
    """ Defines clip module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.CMP * 2
        self.ratio = 1


class Relu6(Clip):
    """ Defines relu6 module hook. """
    pass


class Exp(PointwiseBase):
    """ Defines exp module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.EXP
        self.ratio = 1


class Log(PointwiseBase):
    """ Defines log module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.LOG
        self.ratio = 1


class CumSum(PointwiseBase):
    """ Defines cumulative_sum module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.ADD
        self.ratio = 1


class Softmax(Exp):
    """ Defines softmax module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.EXP + MACS.DIV
        self.ratio = 1


class Sigmoid(Exp):
    """ Defines sigmoid module hook. """
    pass


class Tanh(PointwiseBase):
    """ Defines tanh module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.EXP
        self.ratio = 2


class Mul(PointwiseBase):
    """ Defines mul module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.MUL

    def inference_shape(self, inputs: list[np.ndarray]):
        return [inputs[0] * inputs[1]]


class InstanceNormalization(PointwiseBase):
    """ Defines instance_normalization module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = 2 * MACS.ADD + MACS.MUL + MACS.DIV


class Sqrt(PointwiseBase):
    """ Defines sqrt module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.SQRT


class Pow(PointwiseBase):
    """ Defines pow module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.POW


class Sin(PointwiseBase):
    """ Defines sin module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.SIN


class Cos(PointwiseBase):
    """ Defines cos module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.COS


class Div(PointwiseBase):
    """ Defines div module hook. """
    
    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.DIV

    def inference_shape(self, inputs: list[np.ndarray]):
        return [inputs[0] / (inputs[1])]


class Range(PointwiseBase):
    """ Defines range module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = 1

    def inference_shape(self, inputs: list[np.ndarray]):
        return [np.arange(inputs[0], inputs[1], inputs[2], dtype=np.float32)]


class Reciprocal(PointwiseBase):
    """ Defines reciprocal module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.DIV


class HardSigmoid(PointwiseBase):
    """ Defines hard_sigmoid module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.MUL + MACS.ADD + MACS.CMP * 2


class LeakyRelu(PointwiseBase):
    """ Defines leaky_relu module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.MUL + MACS.CMP


import numpy as np
import math

from .common.utilities import *
from .common.functions import *
from .common.constants import MACS, ZERO_OP

from .pointwise import PointwiseBase
from .elementwise import (Conv, Gemm)


class DequantizeLinear(PointwiseBase):
    """ Defines dequantize_linear module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.MUL


class QuantizeLinear(PointwiseBase):
    """ Defines quanzize_linear module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        self.operation_mac_count = MACS.MUL


class QuantizeLinearMatMul(Gemm):
    """ Defines quanzize_linear_mat_mul module hook. """

    def __init__(self, proto):
        super().__init__(proto)
        attribute_names = ['transA', 'transB']
        self.transA = None
        self.transB = None
        set_attributes(proto.attribute, attribute_names, self)

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        x_shape = inputs[0].shape
        weight_shape = inputs[3].shape
        macs = construct_volume(x_shape)
        if self.__class__ == Gemm:
            macs *= weight_shape[0]
        else:
            macs *= weight_shape[-1]
        return (macs, ZERO_OP)

    def inference_shape(self, inputs: list[np.ndarray]):
        x = inputs[0]
        w = inputs[3]

        if self.__class__ == Gemm:
            if self.transA is not None and self.transA > 0:
                x_shape = x.shape[::-1]
            else:
                x_shape = x.shape
            if self.transB is not None and self.transB > 0:
                yshape = x_shape[:-1] + (w.shape[-2],)
            else:
                yshape = x_shape[:-1] + (w.shape[-1],)
        else:
            yshape = x.shape[:-1] + (w.shape[-1],)

        return [construct_ndarray(yshape, dtype=np.float32)]


class QuantizeLinearConv(Conv):
    """ Defines quantize_linear_conv module hook. """

    def profile(self, inputs: list[np.ndarray], outputs: list[np.ndarray]):
        if self.dimensions["output"] == 1:
            kernel_shape = inputs[3].shape
            params = construct_volume(kernel_shape)
            if self.dimensions["inputs"] == 3:
                params += kernel_shape[0]

            if len(kernel_shape) > 3:
                output_volume = construct_volume(outputs[0].shape)
                macs = output_volume * kernel_shape[1] * kernel_shape[2] * kernel_shape[3]
            elif len(kernel_shape) == 3:
                output_volume = construct_volume(outputs[0].shape)
                macs = output_volume * kernel_shape[1] * kernel_shape[2]

            if self.dimensions["inputs"] == 9:
                macs += (output_volume * MACS.ADD)

        return (macs, params)

    def inference_shape(self, inputs: list[np.ndarray]):
        x_tensor = inputs[0]
        w_tensor = inputs[3]
        x_shape = x_tensor.shape
        w_shape = w_tensor.shape

        shape = []

        if self.auto_padding is None or self.auto_padding == b'NOTSET':
            if len(x_shape) == 4:
                output_h = calculate_conv_output_shape(
                    x_shape[2], 
                    self.paddings[0] + self.paddings[2], 
                    w_shape[2], 
                    self.strides[0], 
                    self.dilations[0]
                )
                output_w = calculate_conv_output_shape(
                    x_shape[3], 
                    self.paddings[1] + self.paddings[3], 
                    w_shape[3], 
                    self.strides[1], 
                    self.dilations[1]
                )
                shape = (x_shape[0], w_shape[0], output_h, output_w)
            elif len(x_shape) == 3:
                output_h = calculate_conv_output_shape(
                    x_shape[2], 
                    self.paddings[0] + self.paddings[1], 
                    w_shape[2], 
                    self.strides[0],
                    self.dilations[0]
                )
                shape = (x_shape[0], w_shape[0], output_h)
        elif self.auto_padding in [b'SAME_LOWER', b'SAME_UPPER']:
            shape = (x_shape[0], w_shape[0], math.ceil(x_shape[2] / self.strides[0]))
            if len(x_shape) == 4:
                shape += (math.ceil(x_shape[3] / self.strides[1]),)
        return [construct_ndarray(shape, dtype=np.float32)]


from setuptools import setup, find_packages
from onnxprofile import __VERSION__, __AUTHOR__, __DESCRIPTION__

long_description = open("README.md", "r").read()
requirements = ["onnx", "numpy", "tabulate"]

setup(
    name="onnxprofile",                    
    version=__VERSION__,                   
    author=__AUTHOR__,                     
    description=__DESCRIPTION__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],                                     
    zip_safe=True,
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'onnxprofile = onnxprofile.__main__:main',
        ]
    },
)


import numpy as np


def get_tensor_shape(x):
    output_shape = []
    for d in x.type.tensor_type.shape.dim:
        assert (d.dim_value != None)
        output_shape.append(d.dim_value)
    return output_shape


def get_initializer_shape(x):
    return list(x.dims)


def construct_volume(shape):
    volume = 1 if len(shape) > 0 else 0
    return volume * np.prod(shape)


def construct_ndarray(shape, dtype=np.float32):
    if dtype == np.float32:
        return np.ones(shape, dtype=dtype)
    elif dtype == np.int64:
        return np.zeros(shape, dtype=dtype)
    else:
        raise NotImplementedError(
            f"dtype: {dtype} currently not implemented"
        )


def validate_ndarray(ndarray):
    if isinstance(ndarray, (list, tuple)) and len(ndarray) == 0 or ndarray is None:
        return False
    elif isinstance(ndarray, np.ndarray):
        return bool(ndarray.size) if construct_volume(ndarray.shape) == 0 else True
    return False


def get_dtype(x):
    if x.data_type == x.FLOAT:
        return np.float32
    elif x.data_type == x.INT64:
        return np.int64
    elif x.data_type == x.INT32:
        return np.int32
    elif x.data_type == x.INT16:
        return np.int16
    elif x.data_type == x.INT8:
        return np.int8
    elif x.data_type == x.UNIT8:
        return np.uint8
    elif x.data_type == x.BOOL:
        return np.bool
    

def transform_to_ndarray(x):
    shape = get_initializer_shape(x)
    dtype = get_dtype(x)

    if x.raw_data != b'':
        return np.frombuffer(x.raw_data, dtype=dtype).reshape(shape)
    arr = np.zeros(shape, dtype=dtype).reshape((-1))
    if dtype == np.float64:
        for i in range(len(x.double_data)):
            arr[i] = x.double_data[i]
    elif dtype == np.float32:
        for i in range(len(x.float_data)):
            arr[i] = x.float_data[i]
    elif dtype == np.int64:
        for i in range(len(x.int64_data)):
            arr[i] = x.int64_data[i]
    elif dtype == np.int32:
        for i in range(len(x.int32_data)):
            arr[i] = x.int32_data[i]
    return arr.reshape(shape)


def get_attribute(a):
    if a.type == a.INTS:
        return list(a.ints)
    elif a.type == a.FLOATS:
        return list(a.floats)
    elif a.type == a.INT:
        return a.i
    elif a.type == a.FLOAT:
        return a.f
    elif a.type == a.STRING:
        return a.s
    elif a.type == a.TENSOR:
        return transform_to_ndarray(a.t)


def set_attributes(attributes, names, obj):
    for a in attributes:
        if a.name in names:
            obj.__setattr__(a.name, get_attribute(a))


def remove_initialisers(model):
    model.graph.ClearField("initializer")


def remove_constants(model):
    valid_nodes = [node for node in model.graph.node if node.op_type != "Constant"]
    model.graph.ClearField("node")
    for node in valid_nodes:
        model.graph.node.append(node)